@@ -2,12 +2,13 @@
 // RUN: %clangxx_asan %s -o %t
 // The memory is released only when the deallocated chunk leaves the quarantine,
 // otherwise the mmap(p, ...) call overwrites the malloc header.
-// RUN: ASAN_OPTIONS=quarantine_size=1 %run %t
+// RUN: ASAN_OPTIONS=quarantine_size_mb=0 %run %t
 
 #include <assert.h>
 #include <string.h>
 #include <sys/mman.h>
 #include <stdlib.h>
+#include <unistd.h>
 
 #ifdef __ANDROID__
 #include <malloc.h>
@@ -23,11 +24,12 @@ void *my_memalign(size_t boundary, size_t size) {
 #endif
 
 int main() {
-  const int kPageSize = 4096;
+  const long kPageSize = sysconf(_SC_PAGESIZE);
   void *p = my_memalign(kPageSize, 1024 * 1024);
   free(p);
 
-  char *q = (char *)mmap(p, kPageSize, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANON | MAP_FIXED, 0, 0);
+  char *q = (char *)mmap(p, kPageSize, PROT_READ | PROT_WRITE,
+                         MAP_PRIVATE | MAP_ANON | MAP_FIXED, -1, 0);
   assert(q == p);
 
   memset(q, 42, kPageSize);
